import os
import runpy

from git import Repo
from semantic_version import Version

from git_semver import get_current_version
from medikit.steps import Step


class PythonVersion(Version):
    def __str__(self):
        version = '%d' % self.major
        if self.minor is not None:
            version = '%s.%d' % (version, self.minor)
        if self.patch is not None:
            version = '%s.%d' % (version, self.patch)

        if self.prerelease or (self.partial and self.prerelease == () and self.build is None):
            version = '%s%s' % (version, '.'.join(self.prerelease))
        if self.build or (self.partial and self.build == ()):
            version = '%s+%s' % (version, '.'.join(self.build))
        return version


class BumpVersion(Step):
    def run(self, meta):
        '''
        git add $VERSION_FILE
        '''
        name = self.exec('python setup.py --name')
        version_file = os.path.join(name.replace('.', '/'), '_version.py')

        if not os.path.exists(version_file):
            raise FileNotFoundError('Cannot find version file for {} (searched in {!r}).'.format(name, version_file))

        repo = Repo()
        for remote in repo.remotes:
            self.logger.info('git fetch {}...'.format(remote))
            remote.fetch(tags=True)

        git_version = get_current_version(repo, Version=PythonVersion)
        if git_version:
            git_version.partial = False
        current_version = PythonVersion.coerce(runpy.run_path(version_file).get('__version__'), partial=True)
        current_version.partial = False

        next_version = None
        while not next_version:
            try:
                print('Current version:', current_version, 'Git version:', git_version)
                next_version = input('Next version? ')
                # todo next patch, etc.
                next_version = PythonVersion.coerce(next_version, partial=True)
                next_version.partial = False
            except ValueError as exc:
                self.logger.error(exc)
                next_version = None

        with open(version_file, 'w+') as f:
            f.write("__version__ = '{}'\n".format(str(next_version)))

        self.exec('git add {}'.format(version_file))

        meta['version'] = str(next_version)

        self.set_complete()



"""
Pipelines are a way to describe a simple step-by-step process, for example the release process.

"""
import datetime
import json
import logging

logger = logging.getLogger(__name__)


class Pipeline:
    """
    Class to configure a pipeline.

    """

    def __init__(self):
        self.steps = []

    def add(self, step):
        self.steps.append(step)
        return self

    def remove(self, identity):
        for i in range(len(self.steps)):
            if identity == get_identity(self.steps[i]):
                del self.steps[i]
                break


def get_identity(step):
    return str(step)


class ConfiguredPipeline:
    """
    Used to actually load run and persist a configured pipeline.
    """

    def __init__(self, name, pipeline):
        self.name = name
        self.steps = pipeline.steps
        self.meta = {'created': str(datetime.datetime.now())}

    def init(self):
        for step in self.steps:
            step.init()

    def next(self):
        for step in self.steps:
            if not step.complete:
                return step
        raise StopIteration('No step left.')

    @property
    def current(self):
        for i, step in enumerate(self.steps):
            if not step.complete:
                return i + 1
        return len(self)

    def __len__(self):
        return len(self.steps)

    def abort(self):
        for step in self.steps:
            step.abort()

    def serialize(self):
        return json.dumps(
            {
                'meta': {
                    **self.meta,
                    'updated': str(datetime.datetime.now()),
                },
                'steps': [[get_identity(step), step.get_state()] for step in self.steps]
            },
            indent=4
        )

    def unserialize(self, serialized):
        serialized = json.loads(serialized)
        self.meta = serialized.get('meta', {})
        steps = serialized.get('steps', [])
        if len(steps) != len(self.steps):
            raise IOError('Invalid pipeline state storage.')
        for (identity, state), step in zip(steps, self.steps):
            if get_identity(step) != identity:
                raise IOError('Mismatch on step identity.')
            step.set_state(state)



import logging
import textwrap

from collections import OrderedDict
from whistle import Event, EventDispatcher

from mondrian import term

logger = logging.getLogger()


class ProjectEvent(Event):
    """
    :attr OrderedDict variables:
    :attr dict files:
    :attr OrderedDict setup:
    """

    def __init__(self, *, config, variables=None, files=None, setup=None):
        """
        :param OrderedDict|NoneType variables:
        """
        self.config = config
        self.variables = OrderedDict(variables or {})
        self.files = dict(files or {})
        self.setup = OrderedDict(setup or {})

        super(ProjectEvent, self).__init__()


class LoggingDispatcher(EventDispatcher):
    logger = logging.getLogger()
    indent_level = 0

    @property
    def indent(self):
        return '  ' * type(self).indent_level

    def dispatch(self, event_id, event=None):
        should_log = not event_id.startswith('medikit.on_file_') or \
                     self.logger.getEffectiveLevel() <= logging.DEBUG
        if should_log:
            self.logger.info(
                self.indent + term.bold('>') +
                ' dispatch ⚡ {} ({})'.format(term.bold(term.blue(event_id)),
                                             type(event or Event).__name__)
            )
        type(self).indent_level += 1
        event = super(LoggingDispatcher, self).dispatch(event_id, event)
        type(self).indent_level -= 1
        if should_log:
            self.logger.info(self.indent + term.bold('<') + ' {}'.format(term.lightblack('dispatched ' + event_id)))
        return event

    def debug(self, feature, *messages):
        return self.logger.debug('   ✔ ' + term.bold(term.green(feature.__shortname__)) + ' '.join(map(str, messages)))

    def info(self, *messages):
        return self.logger.info(self.indent + term.lightblack('∙') + ' ' + ' '.join(map(str, messages)))


def subscribe(event_id, priority=0):
    """
    Lazy event subscription. Will need to be attached to an event dispatcher using ``attach_subscriptions``

    :param str event_id:
    :param int priority:
    :return:
    """

    def wrapper(f):
        try:
            getattr(f, '__subscriptions__')
        except AttributeError as e:
            f.__subscriptions__ = {}
            f.__doc__ = '' if f.__doc__ is None else textwrap.dedent(f.__doc__).strip()
            if f.__doc__:
                f.__doc__ += '\n'

        f.__subscriptions__[event_id] = priority
        f.__doc__ += '\nListens to ``{}`` event *(priority: {})*'.format(event_id, priority)

        return f

    return wrapper


def attach_subscriptions(obj, dispatcher):
    """
    Attach subscriptions to an actual event dispatcher.

    :param object obj:
    :param EventDispatcher dispatcher:
    :return:
    """
    for k in dir(obj):
        f = getattr(obj, k)
        if k[0] != '_' and callable(f) and hasattr(f, '__subscriptions__'):
            for event_id, priority in f.__subscriptions__.items():
                dispatcher.add_listener(event_id, f, priority=priority)



from mondrian import term

from medikit.commands.base import Command
from medikit.commands.utils import _read_configuration
from medikit.config.loader import load_feature_extensions
from medikit.events import LoggingDispatcher, ProjectEvent


class UpdateCommand(Command):
    @staticmethod
    def handle(config_filename, **kwargs):
        import logging
        logger = logging.getLogger()

        dispatcher = LoggingDispatcher()

        variables, features, files, config = _read_configuration(dispatcher, config_filename)

        feature_instances = {}
        logger.info(
            'Updating {} with {} features'.format(
                term.bold(config['python'].get('name')),
                ', '.join(term.bold(term.green(feature_name)) for feature_name in sorted(features))
            )
        )

        all_features = load_feature_extensions()

        sorted_features = sorted(features)  # sort to have a predictable display order
        for feature_name in sorted_features:
            logger.debug('Initializing feature {}...'.format(term.bold(term.green(feature_name))))
            try:
                feature = all_features[feature_name]
            except KeyError as exc:
                logger.exception('Feature "{}" not found.'.format(feature_name))

            if feature:
                feature_instances[feature_name] = feature(dispatcher)

                for req in feature_instances[feature_name].requires:
                    if not req in sorted_features:
                        raise RuntimeError('Unmet dependency: {} requires {}.'.format(feature_name, req))

                for con in feature_instances[feature_name].conflicts:
                    if con in sorted_features:
                        raise RuntimeError('Conflicting dependency: {} conflicts with {}.'.format(con, feature_name))
            else:
                raise RuntimeError('Required feature {} not found.'.format(feature_name))

        event = ProjectEvent(config=config)
        event.variables, event.files = variables, files

        # todo: add listener dump list in debug/verbose mode ?

        event = dispatcher.dispatch('medikit.on_start', event)

        dispatcher.dispatch('medikit.on_end', event)

        logger.info('Done.')



"""
Webpack support.

This feature is experimental and as though it may work for you, that's not a guarantee. Please use with care.

"""

from medikit.events import subscribe
from . import Feature


class WebpackFeature(Feature):
    requires = {'nodejs'}

    @subscribe('medikit.feature.make.on_generate')
    def on_make_generate(self, event):
        event.makefile.get_target('install').install.append('$(YARN) --version')


__feature__ = WebpackFeature



import os
from contextlib import contextmanager

from medikit.commands.base import Command
from medikit.events import LoggingDispatcher
from medikit.feature import ProjectInitializer


@contextmanager
def _change_working_directory(path):
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


class InitCommand(Command):
    def add_arguments(self, parser):
        parser.add_argument('target')
        parser.add_argument('--name')
        parser.add_argument('--description')
        parser.add_argument('--license')
        parser.add_argument('--feature', '-f', action='append', dest='features')

    @staticmethod
    def handle(config_filename, **options):
        if os.path.exists(config_filename):
            raise IOError(
                'No config should be present in current directory to initialize (found {})'.format(config_filename)
            )

        config_dirname = os.path.dirname(config_filename)
        if not os.path.exists(config_dirname):
            os.makedirs(config_dirname)

        # Fast and dirty implementation
        # TODO
        # - input validation
        # - getting input from env/git conf (author...),
        # - dispatching something in selected features so maybe they can suggest deps
        # - deps selection
        # - ...
        with _change_working_directory(config_dirname):
            dispatcher = LoggingDispatcher()
            initializer = ProjectInitializer(dispatcher, options)
            initializer.execute()
            from medikit.commands import UpdateCommand
            return UpdateCommand.handle(config_filename)



"""
Lint your python code with pylint.

This feature may be outdated.

"""
from medikit.events import subscribe

from . import Feature, SUPPORT_PRIORITY


class PylintFeature(Feature):
    requires = {'python'}

    @subscribe('medikit.feature.python.on_generate')
    def on_python_generate(self, event):
        event.config['python'].add_requirements(dev=['pylint ~=1.8'])

    @subscribe('medikit.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile
        makefile.add_target(
            'lint',
            '''
            $(PYTHON_DIRNAME)/pylint --py3k $(PACKAGE) -f html > pylint.html
        '''.format(name=event.package_name),
            deps=('install-dev', ),
            phony=True
        )


__feature__ = PylintFeature



from contextlib import contextmanager

from stevedore import ExtensionManager

from medikit.pipeline import Pipeline


class ConfigurationRegistry():
    def __init__(self):
        self._configs = {}
        self._features = {}
        self._pipelines = {}

        def register_feature(ext):
            self._features[ext.name] = ext.plugin

        mgr = ExtensionManager(namespace='medikit.feature')
        mgr.map(register_feature)

    def __getitem__(self, item):
        return self._configs[item]

    def keys(self):
        return self._configs.keys()

    def require(self, *args):
        if len(args) == 1:
            return self._require(args[0])
        elif len(args) > 1:
            return tuple(map(self._require, args))
        raise ValueError('Empty.')

    @contextmanager
    def pipeline(self, name):
        if not name in self._pipelines:
            self._pipelines[name] = Pipeline()
        yield self._pipelines[name]

    @property
    def pipelines(self):
        return self._pipelines

    def _require(self, name):
        if not name in self._features:
            raise ValueError('Unknown feature {!r}.'.format(name))

        if name not in self._configs:
            self._configs[name] = self._features[name].Config()

        return self._configs[name]


"""
Git version control system support.

"""

import os

from medikit.events import subscribe
from medikit.feature import ABSOLUTE_PRIORITY, Feature


class GitConfig(Feature.Config):
    __usage__ = """
    
    Currently, **this feature is required for medikit to work**.
    
    There are plans to add the ability to disable it, but it's not a priority.
    
    """


class GitFeature(Feature):
    Config = GitConfig

    @subscribe('medikit.on_start', priority=ABSOLUTE_PRIORITY)
    def on_start(self, event):
        if not os.path.exists('.git'):
            self.dispatcher.info('git', 'Creating git repository...')
            os.system('git init')
            os.system('git add Projectfile')
            os.system('git commit -m "Project initialized using Medikit."')

    @subscribe('medikit.on_end')
    def on_end(self, event):
        self.render_file_inline(
            '.gitignore', '''
            *.egg-info
            *.iml
            *.pyc
            *.swp
            /.cache
            /.coverage
            /.idea
            /.python*-*
            /build
            /dist
            /htmlcov
            /pylint.html
        ''', event.variables
        )

    @subscribe('medikit.on_file_closed')
    def on_file_change(self, event):
        os.system('git add {}'.format(event.filename))

    @subscribe('medikit.feature.make.on_generate', priority=ABSOLUTE_PRIORITY + 1)
    def on_make_generate(self, event):
        event.makefile['VERSION'] = "$(shell git describe 2>/dev/null || git rev-parse --short HEAD)"


__feature__ = GitFeature



import os
from collections import OrderedDict

from medikit.config import read_configuration
from medikit.settings import DEFAULT_FILES, DEFAULT_FEATURES


def _read_configuration(dispatcher, config_filename):
    """
    Prepare the python context and delegate to the real configuration reader (see config.py)

    :param EventDispatcher dispatcher:
    :return tuple: (variables, features, files, config)
    """
    if not os.path.exists(config_filename):
        raise IOError('Could not find project description file (looked in {})'.format(config_filename))

    variables = OrderedDict()

    files = {filename: '' for filename in DEFAULT_FILES}
    features = set(DEFAULT_FEATURES)

    return read_configuration(dispatcher, config_filename, variables, features, files)


# This is stolen from honcho, probably needs refactoring / simplification for our usage.
import datetime
import os
import signal
import subprocess

import sys
from collections import namedtuple

Message = namedtuple("Message", "type data time name colour")


class Process(object):
    def __init__(self, cmd, name=None, colour=None, quiet=False, env=None, cwd=None):
        self.cmd = cmd
        self.colour = colour
        self.quiet = quiet
        self.name = name
        self.env = os.environ.copy() if env is None else env
        self.cwd = cwd

        self._child = None
        self._child_ctor = Popen

    def run(self, events=None, ignore_signals=False):
        self._events = events
        self._child = self._child_ctor(self.cmd, env=self.env, cwd=self.cwd)
        self._send_message({'pid': self._child.pid}, type='start')

        # Don't pay attention to SIGINT/SIGTERM. The process itself is
        # considered unkillable, and will only exit when its child (the shell
        # running the Procfile process) exits.
        if ignore_signals:
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)

        for line in iter(self._child.stdout.readline, b''):
            if not self.quiet:
                self._send_message(line)
        self._child.stdout.close()
        self._child.wait()

        self._send_message({'returncode': self._child.returncode}, type='stop')

    def _send_message(self, data, type='line'):
        if self._events is not None:
            self._events.put(
                Message(type=type, data=data, time=datetime.datetime.now(), name=self.name, colour=self.colour)
            )


ON_WINDOWS = 'win32' in str(sys.platform).lower()


class Popen(subprocess.Popen):
    def __init__(self, cmd, **kwargs):
        start_new_session = kwargs.pop('start_new_session', True)
        options = {
            'stdout': subprocess.PIPE,
            'stderr': subprocess.STDOUT,
            'shell': True,
            'bufsize': 1,
            'close_fds': not ON_WINDOWS,
        }
        options.update(**kwargs)

        if ON_WINDOWS:
            # MSDN reference:
            #   http://msdn.microsoft.com/en-us/library/windows/desktop/ms684863%28v=vs.85%29.aspx
            create_new_process_group = 0x00000200
            detached_process = 0x00000008
            options.update(creationflags=detached_process | create_new_process_group)
        elif start_new_session:
            if sys.version_info < (3, 2):
                options.update(preexec_fn=os.setsid)
            else:
                options.update(start_new_session=True)

        super(Popen, self).__init__(cmd, **options)



"""
Adds the pytest testing framework to your project.

"""

import os

from medikit.events import subscribe
from . import Feature, SUPPORT_PRIORITY


class PytestFeature(Feature):
    # TODO: http://docs.pytest.org/en/latest/goodpractices.html#integrating-with-setuptools-python-setup-py-test-pytest-runner

    requires = {'python'}

    @subscribe('medikit.feature.python.on_generate')
    def on_python_generate(self, event):
        event.config['python'].add_requirements(
            dev=['coverage ~=4.4', 'pytest ~=3.4', 'pytest-cov ~=2.5', 'pytest-sugar ~=0.9.1']
        )

    @subscribe('medikit.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile
        makefile['PYTEST'] = '$(PYTHON_DIRNAME)/pytest'
        makefile['PYTEST_OPTIONS'] = '--capture=no --cov=$(PACKAGE) --cov-report html'.format(
            path=event.package_name.replace('.', os.sep)
        )
        makefile.add_target(
            'test',
            '''
            $(PYTEST) $(PYTEST_OPTIONS) tests
        ''',
            deps=('install-dev', ),
            phony=True,
            doc='Runs the test suite.'
        )

    @subscribe('medikit.on_start', priority=SUPPORT_PRIORITY)
    def on_start(self, event):
        tests_dir = 'tests'
        if not os.path.exists(tests_dir):
            os.makedirs(tests_dir)

        gitkeep_file = os.path.join(tests_dir, '.gitkeep')

        if not os.path.exists(gitkeep_file):
            self.render_empty_files(gitkeep_file)

        self.render_file('.coveragerc', 'pytest/coveragerc.j2')
        self.render_file('.travis.yml', 'pytest/travis.yml.j2')


__feature__ = PytestFeature



"""
The “python” feature is the base feature required for all python projects.

Medikit only supports python 3.5+ projects, which we believe is a future proof choice.

Overview
::::::::

.. code-block:: python

    from medikit import require

    # load python feature (idempotent).
    python = require('python')

    # configure our package.
    python.setup(
        name = 'awesome-library',
        author = 'Chuck Norris',
    )

    # add base and extras requirements, with "loose" versionning (the frozen version fits better in requirements*.txt)
    python.add_requirements(
        'django',
        dev=[
            'pytest',
        ],
    )

"""

import itertools
import os
import tempfile
from getpass import getuser

from pip._vendor.distlib.util import parse_requirement
from pip.req import parse_requirements
from piptools.repositories import PyPIRepository
from piptools.resolver import Resolver
from piptools.scripts.compile import get_pip_command
from piptools.utils import format_requirement

from medikit.events import subscribe
from medikit.feature import Feature, ABSOLUTE_PRIORITY
from medikit.feature.make import which
from medikit.utils import get_override_warning_banner


def _normalize_requirement(req):
    bits = req.requirement.split()
    if req.extras:
        bits = [bits[0] + '[{}]'.format(','.join(req.extras))] + bits[1:]
    return ' '.join(bits)


class PythonConfig(Feature.Config):
    """ Configuration API for the «python» feature. """

    on_generate = __name__ + '.on_generate'

    __usage__ = """
    
    The python features makes your package real (at least if it uses python). Medikit was written originally for python,
    and although it's not completely true anymore, a lot of features depends on this.
    
    Setup
    -----
    
    You can define the setuptools' `setup(...)` arguments using `python.setup(...)`:
    
    .. code-block:: python
    
        python.setup(
            name='medikit',
            description='Opinionated python 3.5+ project management.',
            license='Apache License, Version 2.0',
            url='https://github.com/python-medikit/medikit',
            download_url='https://github.com/python-medikit/medikit/tarball/{version}',
            author='Romain Dorgueil',
            author_email='romain@dorgueil.net',
            entry_points={
                'console_scripts': ['medikit=medikit.__main__:main'],
            }
        )
        
    This is required for any python package.
    
    Requirements
    ------------
    
    Requirements are managed using two different mechanisms:
    
    * The `setup.py` file, autogenerated and overriden by `medikit`, will contain the "loose" requirements as you define
      them. You're encouraged to use "~x.y.z" or "~x.y" versions. You should use each versions ("==x.y.z") only in case
      you're relying on a package you never want to update.
    * The `requirements*.txt` files will contain frozen version numbers. Those requirements will be commited, and you
      can ensure the reproducibility of your installs by using `pip install -r requirements.txt` instead of
      `python setup.py install`.
      
    In `medikit`, we call what is present in setup.py "constraints", and what is in `requirements*.txt` files
    "requirements".
    
    Let's see how we can set them:
    
    .. code-block:: python
    
        python.add_requirements(
            "requests ~2.18"
        )
        
    This will set a constraint on any semver compatible requests version, and update the requirement to latest requests
    version, compatible with the constraint (as of writing, 2.18.4).
    
    It means that if you run `make install`, `python setup.py install` or `pip install -e .`, requests will only be
    downloaded and installed if there is no installation complying to the constraint in your current env. This is very
    handy if you have local, editable packages that you want to use instead of PyPI versions.
    
    It also means that when you run `pip install -r requirements.txt`, you'll get requests 2.18.4 even if a new version
    was released.
    
    If you want to upgrade to the new released version, use `make update-requirements`, review the git changes
    (`git diff --cached`), test your software with the new version and eventually (git) commit to this dependency update.
    
    Constraints
    -----------
    
    Sometimes, you want a dependency to only be a constraint, and not a frozen requirement.
    
    .. code-block:: python
    
        python.add_constraints(
            "certifi ~2018,<2019"
        )
        
    This will ensure that your env contains "certifi", a version released in the 2018 year, but also says you don't care
    which one.
    
    This is an advanced feature that you should only use if you really know what you're doing, otherwise, use a
    requirement (reproducibility of installs is gold).
    
    Extras
    ------
    
    You can create as much "extras" as you want.
    
    As a default, medikit will create a "dev" extra, but you can add whatever you need:
    
    .. code-block:: python
    
        python.add_requirements(
            sql=[
                'sqlalchemy ~=1.2.5',
            ]
        )
    
    The same works with constraints, of course.
    
    Changing package generation behaviour
    -------------------------------------
    
    Medikit creates the necessary directory structure for your package, named after your package name defined in the
    `python.setup()` call.
    
    If you don't want medikit to create this directory structure:
    
    .. code-block:: python
    
        python.create_packages = False
    
    Medikit also considers you'll need a version number tracking mechanism for your project. It creates a `_version.py` 
    file in your package's root directory. To override this file's name:
    
    .. code-block:: python
    
        python.version_file = 'my_version.py'
    
    """

    def __init__(self):
        self._setup = {}
        self._requirements = {None: {}, 'dev': {}}
        self._constraints = {}
        self._version_file = None
        self._create_packages = True

    @property
    def package_dir(self):
        return '/'.join(self.get('name').split('.'))

    @property
    def version_file(self):
        return self._version_file or os.path.join(self.package_dir, '_version.py')

    @version_file.setter
    def version_file(self, value):
        self._version_file = value

    @property
    def create_packages(self):
        return self._create_packages

    @create_packages.setter
    def create_packages(self, value):
        self._create_packages = bool(value)

    def get(self, item):
        if item == 'install_requires':
            return list(self.get_requirements(with_constraints=True))
        if item == 'extras_require':
            return {
                extra: list(self.get_requirements(extra=extra, with_constraints=True))
                for extra in self.get_extras()
            }
        return self._setup.get(item)

    def get_init_files(self):
        # Explode package name so we know which python packages are namespaced and
        # which are not.
        package_bits = self.get('name').split('.')
        if len(package_bits) > 1:
            # XXX SIDE EFFECT !!!
            self['namespace_packages'] = []
            for i in range(1, len(package_bits)):
                self['namespace_packages'].append('.'.join(package_bits[0:i]))

        for i in range(1, len(package_bits) + 1):
            _bits = package_bits[0:i]
            package_dir = os.path.join(*_bits)
            package_init_file = os.path.join(package_dir, '__init__.py')
            is_namespace = i < len(package_bits)
            yield package_dir, package_init_file, {'is_namespace': is_namespace}

    def __getitem__(self, item):
        return self.get(item)

    def __setitem__(self, key, value):
        self.setup(**{key: value})

    def __contains__(self, item):
        return bool(self.get(item))

    def add_constraints(self, *reqs, **kwargs):
        self.__add_constraints(reqs)
        for extra, reqs in kwargs.items():
            extra = extra.replace('_', '-')
            self.__add_constraints(reqs, extra=extra)
        return self

    def add_requirements(self, *reqs, **kwargs):
        self.__add_requirements(reqs)
        for extra, reqs in kwargs.items():
            extra = extra.replace('_', '-')
            self.__add_requirements(reqs, extra=extra)
        return self

    def get_extras(self):
        return sorted(filter(None, self._requirements.keys()))

    def get_constraints(self, extra=None):
        for name, req in sorted(self._constraints.get(extra, {}).items()):
            yield _normalize_requirement(req)

    def get_requirements(self, extra=None, with_constraints=False):
        if with_constraints:
            yield from self.get_constraints(extra=extra)
        for name, req in sorted(self._requirements[extra].items()):
            yield _normalize_requirement(req)

    def setup(self, **kwargs):
        self._setup.update(kwargs)
        return self

    def get_setup(self):
        return self._setup

    def __add_constraints(self, reqs, extra=None):
        if len(reqs):
            if extra not in self._constraints:
                self._constraints[extra] = {}
        for req in reqs:
            req = parse_requirement(req)
            if req.name in self._constraints[extra]:
                raise ValueError('Duplicate constraint for {}.'.format(req.name))
            self._constraints[extra][req.name] = req

    def __add_requirements(self, reqs, extra=None):
        if len(reqs):
            if extra not in self._requirements:
                self._requirements[extra] = {}
        for req in reqs:
            req = parse_requirement(req)
            if req.name in self._requirements[extra]:
                raise ValueError('Duplicate requirement for {}.'.format(req.name))
            self._requirements[extra][req.name] = req


class PythonFeature(Feature):
    """
    Adds the requirements/requirements-dev logic, virtualenv support, setup.py generation, a few metadata files used by
    setuptools...

    """

    requires = {'make'}

    Config = PythonConfig

    @subscribe('medikit.feature.make.on_generate', priority=ABSOLUTE_PRIORITY)
    def on_make_generate(self, event):
        """
        **Environment variables**

        - ``PACKAGE``
        - ``PYTHON``
        - ``PYTHON_BASENAME``
        - ``PYTHON_DIR``
        - ``PYTHON_REQUIREMENTS_FILE``
        - ``PYTHON_REQUIREMENTS_DEV_FILE``
        
        **Shortcuts**
        - ``PIP``
        - ``PIP_INSTALL_OPTIONS``

        **Make targets**

        - ``install``
        - ``install-dev``

        :param MakefileEvent event:

        """
        # Python related environment
        event.makefile.updateleft(
            (
                'PACKAGE',
                event.package_name,
            ),
            (
                'PYTHON',
                which('python'),
            ),
            (
                'PYTHON_BASENAME',
                '$(shell basename $(PYTHON))',
            ),
            (
                'PYTHON_DIRNAME',
                '$(shell dirname $(PYTHON))',
            ),
            (
                'PYTHON_REQUIREMENTS_FILE',
                'requirements.txt',
            ),
            (
                'PYTHON_REQUIREMENTS_DEV_FILE',
                'requirements-dev.txt',
            ),
        )

        event.makefile['PIP'] = '$(PYTHON_DIRNAME)/pip'
        event.makefile['PIP_INSTALL_OPTIONS'] = ''

        event.makefile.get_target('install').install = [
            '$(PIP) install $(PIP_INSTALL_OPTIONS) -U pip wheel',
            '$(PIP) install $(PIP_INSTALL_OPTIONS) -U -r $(PYTHON_REQUIREMENTS_FILE)',
        ]

        event.makefile.get_target('install-dev').install = [
            '$(PIP) install $(PIP_INSTALL_OPTIONS) -U pip wheel',
            '$(PIP) install $(PIP_INSTALL_OPTIONS) -U -r $(PYTHON_REQUIREMENTS_DEV_FILE)',
        ]

    @subscribe('medikit.on_start', priority=ABSOLUTE_PRIORITY)
    def on_start(self, event):
        """
        **Events**

        - ``medikit.feature.python.on_generate`` (with the same ``ProjectEvent`` we got, todo: why not deprecate
          it in favor of higher priority medikit.on_start?)

        **Files**

        - ``<yourpackage>/__init__.py``
        - ``MANIFEST.in``
        - ``README.rst``
        - ``classifiers.txt``
        - ``requirements-dev.txt``
        - ``requirements.txt``
        - ``setup.cfg``
        - ``setup.py`` (overwritten)

        :param ProjectEvent event:
        """
        self.dispatcher.dispatch(__name__ + '.on_generate', event)

        # Our config object
        python_config = event.config['python']

        # Some metadata that will prove useful.
        self.render_empty_files('classifiers.txt', 'README.rst')
        self.render_file_inline('MANIFEST.in', 'include *.txt')
        self.render_file_inline(
            'setup.cfg', '''
                    [bdist_wheel]
                    # This flag says that the code is written to work on both Python 2 and Python
                    # 3. If at all possible, it is good practice to do this. If you cannot, you
                    # will need to generate wheels for each Python version that you support.
                    universal=1
                    
                    [metadata]
                    description-file = README.rst
                '''
        )

        # XXX there is an important side effect in get_init_files that defines namespace packages for later use in
        # setup.py rendering. Even if we do not want to create the packages here, it is important to make the call
        # anyway.
        for dirname, filename, context in python_config.get_init_files():
            if python_config.create_packages:
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                self.render_file(filename, 'python/package_init.py.j2', context)

        # render version file, try to import version from version.txt if available.
        try:
            with open('version.txt') as f:
                version = f.read().strip()
        except IOError:
            version = '0.0.0'
        self.render_file_inline(python_config.version_file, "__version__ = '{}'".format(version))

        setup = python_config.get_setup()

        context = {
            'url': setup.pop('url', 'http://example.com/'),
            'download_url': setup.pop('download_url', 'http://example.com/'),
        }

        for k, v in context.items():
            context[k] = context[k].format(
                name=setup['name'],
                user=getuser(),
                version='{version}',
            )

        context.update(
            {
                'entry_points': setup.pop('entry_points', {}),
                'extras_require': python_config.get('extras_require'),
                'install_requires': python_config.get('install_requires'),
                'python': python_config,
                'setup': setup,
                'banner': get_override_warning_banner(),
            }
        )

        # Render (with overwriting) the allmighty setup.py
        self.render_file('setup.py', 'python/setup.py.j2', context, override=True)

    @subscribe('medikit.on_end', priority=ABSOLUTE_PRIORITY)
    def on_end(self, event):
        # Our config object
        python_config = event.config['python']

        # Pip / PyPI
        pip_command = get_pip_command()
        pip_options, _ = pip_command.parse_args([])
        session = pip_command._build_session(pip_options)
        repository = PyPIRepository(pip_options, session)

        for extra in itertools.chain((None, ), python_config.get_extras()):
            tmpfile = tempfile.NamedTemporaryFile(mode='wt', delete=False)
            if extra:
                tmpfile.write('\n'.join(python_config.get_requirements(extra=extra)))
            else:
                tmpfile.write('\n'.join(python_config.get_requirements()))
            tmpfile.flush()
            constraints = list(
                parse_requirements(
                    tmpfile.name, finder=repository.finder, session=repository.session, options=pip_options
                )
            )
            resolver = Resolver(constraints, repository, prereleases=False, clear_caches=False, allow_unsafe=False)

            self.render_file_inline(
                'requirements{}.txt'.format('-' + extra if extra else ''),
                '\n'.join(
                    (
                        '-e .{}'.format('[' + extra + ']' if extra else ''),
                        *(('-r requirements.txt', ) if extra else ()),
                        *sorted(
                            format_requirement(req) for req in resolver.resolve(max_rounds=10)
                            if req.name != python_config.get('name')
                        ),
                    )
                )
            )


__feature__ = PythonFeature



"""
GNU Make support.

"""

import itertools
import textwrap
from collections import deque

from whistle import Event

from medikit.events import subscribe
from medikit.feature import Feature, HIGH_PRIORITY
from medikit.structs import Script
from medikit.utils import get_override_warning_banner


def which(cmd):
    return '$(shell which {cmd} || echo {cmd})'.format(cmd=cmd)

class Makefile(object):
    @property
    def targets(self):
        for key in self._target_order:
            yield key, self._target_values[key]

    @property
    def environ(self):
        return self._env_values

    def __init__(self):
        self._env_order, self._env_values, self._env_assignment_operators = deque(), {}, {}
        self._target_order, self._target_values = deque(), {}
        self.phony = set()

    def __delitem__(self, key):
        self._env_order.remove(key)
        del self._env_values[key]

    def __getitem__(self, item):
        return self._env_values[item]

    def __setitem__(self, key, value):
        self._env_values[key] = value
        if not key in self._env_order:
            self._env_order.append(key)

    def __iter__(self):
        for key in self._env_order:
            yield key, self._env_values[key]

    def __len__(self):
        return len(self._env_order)

    def __str__(self):
        content = [
            get_override_warning_banner(),
            '',
        ]

        if len(self):
            for k, v in self:
                v = textwrap.dedent(str(v)).strip()
                v = v.replace('\n', ' \\\n' + ' ' * (len(k) + 4))
                content.append('{} {} {}'.format(k, self._env_assignment_operators.get(k, '?='), v))
            content.append('')

        if len(self.phony):
            content.append('.PHONY: ' + ' '.join(sorted(self.phony)))
            content.append('')

        for target, details in self.targets:
            deps, rule, doc = details
            content.append(
                '{}: {}  ## {}'.format(target, ' '.join(deps),
                                       doc.replace('\n', ' ') if doc else '').strip()
            )

            script = textwrap.dedent(str(rule)).strip()

            for line in script.split('\n'):
                content.append('\t' + line)

            content.append('')

        return '\n'.join(content)

    def add_target(self, target, rule, deps=None, phony=False, first=False, doc=None):
        if target in self._target_order:
            raise RuntimeError('Duplicate definition for make target «{}».'.format(target))

        if isinstance(rule, str):
            rule = Script(rule)

        self._target_values[target] = (
            deps or list(),
            rule,
            textwrap.dedent(doc or '').strip(),
        )
        self._target_order.appendleft(target) if first else self._target_order.append(target)

        if phony:
            self.phony.add(target)

    def get_target(self, target):
        return self._target_values[target][1]

    def set_deps(self, target, deps=None):
        self._target_values[target] = (
            deps or list(),
            self._target_values[target][1],
            self._target_values[target][2],
        )

    def set_assignment_operator(self, key, value):
        assert value in ('?=', '=', '+=', ':=', '::=', '!='), 'Invalid operator'
        self._env_assignment_operators[key] = value

    def setleft(self, key, value):
        self._env_values[key] = value
        if not key in self._env_order:
            self._env_order.appendleft(key)

    def updateleft(self, *lst):
        for key, value in reversed(lst):
            self.setleft(key, value)


class MakefileEvent(Event):
    def __init__(self, package_name, makefile, config):
        self.package_name = package_name
        self.makefile = makefile
        self.config = config
        super(MakefileEvent, self).__init__()


class InstallScript(Script):
    def __init__(self, script=None):
        super(InstallScript, self).__init__(script)

        self.before_install = []
        self.install = self.script
        self.after_install = []

    def __iter__(self):
        yield 'if [ -z "$(QUICK)" ]; then \\'
        for line in map(
                lambda x: '    {} ; \\'.format(x),
                itertools.chain(self.before_install, self.install, self.after_install)
        ):
            yield line
        yield 'fi'


class CleanScript(Script):
    remove = [
        'build',
        'dist',
        '*.egg-info',
    ]

    def __iter__(self):
        yield 'rm -rf {}'.format(' '.join(self.remove))


class MakeConfig(Feature.Config):
    __usage__ = """
    
    Currently, **this feature is required for medikit to work**.
    
    Makefile generation and management is quite central to medikit, and it's one of the strongest opinionated choices
    made by the tool.
    
    .. note:: `Makefile` will be overriden on each `medikit update` run! See below how to customize it.
    
    Everything out of the project's world is managed by a single, central Makefile, that contains all the external
    entrypoints to your package.
    
    By default, it only contains the following targets (a.k.a "tasks"):
    
    * install
    * install-dev
    * update
    * update-requirements
    
    This is highly extendable, and about all other features will add their own targets using listeners to the
    :obj:`MakeConfig.on_generate` event.
    
    Default targets
    ---------------
    
    Install
    .......
    
    The `make install` command will try its best to install your package in your current system environment. For python
    projects, it work with the system python, but you're highly encouraged to use a virtual environment (using
    :mod:`virtualenv` or :mod:`venv` module).
    
    .. code-block:: shel-session
    
        $ make install
        
    Install Dev
    ...........
    
    The `make install-dev` command works like `make install`, but adds the `dev` extra.
    
    The `dev` extra is a convention medikit takes to group all dependencies that are only required to actual hack on
    your project, and that won't be necessary in production / runtime environments.
    
    For python projects, it maps to an "extra", as defined by setuptools. For Node.js projects, it will use the
    "devDependencies".
    
    .. code-block:: shell-session
    
        $ make install-dev
        
    Update
    ......
    
    This is a shortcut to `medikit update`, with a preliminary dependency check on `medikit`.
    
    As you may have noticed, `medikit` is never added as a dependency to your project, so this task will ensure it's
    installed before running.
    
    .. code-block:: shell-session
    
        $ make update
        
    Update Requirements
    ...................
    
    The `make update-requirements` command works like `make update`, but forces the regeneration of `requirements*.txt`
    files.
    
    For security reasons, `medikit` never updates your requirements if they are already frozen in requirements files
    (you would not want a requirement to increment version without notice).
    
    This task is here so you can explicitely update your requirements frozen versions, according to the constraints
    you defined in the `Projectfile`.
    
    .. code-block:: shell-session
    
        $ make update-requirements
        
    Customize your Makefile
    -----------------------
    
    To customize the generated `Makefile`, you can use the same event mechanism that is used by `medikit` features,
    directly from within your `Projectfile`.
    
    Add a target
    ............
    
    .. code-block:: python
    
        from medikit import listen
        
        @listen(make.on_generate)
        def on_make_generate(event):
            event.makefile.add_target('foo', '''
                    echo "Foo!"
                ''', deps=('install', ), phony=True, doc='So foo...'
            )
            
        This is pretty self-explanatory, but let's detail:
        
        * "foo" is the target name (you'll be able to run `make foo`)
        * This target will run `echo "Foo!"`
        * It depends on the `install` target, that needs to be satisfied (install being "phony", it will be run
          every time).
        * This task is "phony", meaning that there will be no `foo` file or directory generated as the output, and thus
          that `make` should consider it's never outdated.
        * If you create non phony targets, they must result in a matching file or directory created.
        * Read more about GNU Make: https://www.gnu.org/software/make/
    
    Change the dependencies of an existing target
    .............................................
    
    .. code-block:: python
    
        from medikit import listen
        
        @listen(make.on_generate)
        def on_make_generate(event):
            event.makefile.set_seps('foo', ('install-dev', ))

    Add (or override) a variable
    ............................
    
    .. code-block:: python
    
        from medikit import listen
        
        @listen(make.on_generate)
        def on_make_generate(event):
            event.makefile['FOO'] = 'Bar'
            
    The user can override `Makefile` variables using your system environment:
    
    .. code-block:: shell-session
    
        $ FOO=loremipsum make foo
        
    To avoid this default behaviour (which is more than ok most of the time), you can change the assignment operator
    used in the makefile.
        
    .. code-block:: python
    
        from medikit import listen
        
        @listen(make.on_generate)
        def on_make_generate(event):
            event.makefile.set_assignment_operator('FOO', ':=')
            
    This is an advanced feature you'll probably never need. You can `read the make variables reference
    <https://www.gnu.org/software/make/manual/html_node/Using-Variables.html#Using-Variables>`_.
    
    """

    on_generate = __name__ + '.on_generate'
    """Happens during the makefile generation."""

    def __init__(self):
        self.include_medikit_targets = True

    def disable_medikit_targets(self):
        self.include_medikit_targets = False


class MakeFeature(Feature):
    Config = MakeConfig

    def configure(self):
        self.makefile = Makefile()

    @subscribe('medikit.on_start', priority=HIGH_PRIORITY)
    def on_start(self, event):
        """
        :param ProjectEvent event:
        """
        for k in event.variables:
            self.makefile[k.upper()] = event.variables[k]

        self.makefile.updateleft((
            'QUICK',
            '',
        ), )

        self.makefile.add_target(
            'install', InstallScript(), phony=True, doc='''Installs the local project dependencies.'''
        )
        self.makefile.add_target(
            'install-dev',
            InstallScript(),
            phony=True,
            doc='''Installs the local project dependencies, including development-only libraries.'''
        )
        self.makefile.add_target('clean', CleanScript(), phony=True, doc='''Cleans up the local mess.''')

        if event.config['make'].include_medikit_targets:
            self.makefile.add_target(
                'update',
                '''
                python -c 'import medikit; print(medikit.__version__)' || pip install medikit;
                $(PYTHON) -m medikit update
            ''',
                phony=True,
                doc='''Update project artifacts using medikit, after installing it eventually.'''
            )

            self.makefile.add_target(
                'update-requirements',
                '''
                rm -rf requirements*.txt
                $(MAKE) update
            ''',
                phony=True,
                doc=
                '''Remove requirements files and update project artifacts using medikit, after installing it eventually.'''
            )

        self.dispatcher.dispatch(
            MakeConfig.on_generate, MakefileEvent(event.config['python'].get('name'), self.makefile, event.config)
        )
        # Recipe courtesy of https://marmelab.com/blog/2016/02/29/auto-documented-makefile.html
        self.makefile.add_target(
            'help',
            r"""
            @echo "Available commands:"
            @echo
            @grep -E '^[a-zA-Z_-]+:.*?##[\s]?.*$$' $(MAKEFILE_LIST) | sort | awk 'BEGIN {FS = ":.*?##"}; {printf "    make \033[36m%-30s\033[0m %s\n", $$1, $$2}'
            @echo
            """,
            phony=True,
            doc='Shows available commands.'
        )
        self.render_file_inline('Makefile', self.makefile.__str__(), override=True)


__feature__ = MakeFeature



import argparse
import logging
from contextlib import contextmanager


class CommandGroup:
    def __init__(self, subparsers):
        self.subparsers = subparsers
        self.children = {}

    def add(self, name, command=None, **kwargs):
        child_parser = self.subparsers.add_parser(name, **kwargs)
        self.children[name] = (command or Command)(child_parser)
        return self.children[name]


class Command:
    @property
    def logger(self):
        try:
            return self._logger
        except AttributeError:
            self._logger = logging.getLogger(type(self).__name__)
            return self._logger

    def __init__(self, parser=None):
        parser = parser or argparse.ArgumentParser()
        parser.set_defaults(_handler=self.handle)
        self.parser = parser
        self.children = {}
        self.add_arguments(parser)

    def add_arguments(self, parser):
        """
        Entry point for subclassed commands to add custom arguments.
        """
        pass

    def handle(self, *args, **options):
        """
        The actual logic of the command. Subclasses must implement this method.
        """
        raise NotImplementedError('Subclasses of {} must provide a handle() method'.format(Command.__name__))

    @contextmanager
    def create_child(self, dest, *, required=False):
        subparsers = self.parser.add_subparsers(dest=dest)
        subparsers.required = required
        self.children[dest] = CommandGroup(subparsers)
        yield self.children[dest]



import shlex
import subprocess
from logging import getLogger


class Step:
    @property
    def logger(self):
        try:
            return self._logger
        except AttributeError:
            self._logger = getLogger(self.__class__.__module__)
            return self._logger

    @property
    def complete(self):
        return self._state.get('complete', False)

    def set_complete(self, value=True):
        self._state['complete'] = bool(value)

    def __init__(self):
        self._state = {}
        self.__args__ = ()

    def __str__(self):
        return '{}({})'.format(type(self).__name__, ', '.join(map(repr, self.__args__)))

    def get_state(self):
        return self._state

    def set_state(self, state):
        self._state = state

    def init(self):
        pass

    def run(self, meta):
        self.set_complete()

    def abort(self):
        pass

    def exec(self, command):
        self.logger.info('Running command %s', command)
        result = subprocess.run(shlex.split(command), stdout=subprocess.PIPE)
        if result.returncode:
            raise RuntimeError(
                '"{command}" exited with status {returncode}.'.format(command=command, returncode=result.returncode)
            )
        return result.stdout.decode('utf-8').strip()



import textwrap


class Script(object):
    """
    Simple structure to hold a shell script for various usage, mainly to use in make targets.

    The goal is to make it flexible, as in alow the user to ammend it after definition.

    """

    def __init__(self, script=None):
        self.set(script)

    def set(self, script=None):
        self.script = self.parse_script(script)

    def prepend(self, script=None):
        """
        Prepend a script to the current script.

        :param script:
        """
        self.script = self.parse_script(script) + self.script

    def append(self, script=None):
        """
        Append a script to the current script.

        :param script:
        """
        self.script = self.script + self.parse_script(script)

    def parse_script(self, script):
        if not script:
            return []
        script = textwrap.dedent(str(script)).strip()
        return script.split('\n')

    def __iter__(self):
        for line in self.script:
            yield line

    def __str__(self):
        return '\n'.join(self.__iter__())



import json
import multiprocessing
import os
from queue import Empty

from git import Repo
from mondrian import term

from medikit.steps import Step
from medikit.steps.utils.process import Process


class System(Step):
    def __init__(self, cmd, *, interractive=False):
        super().__init__()
        self.cmd = cmd
        self.interractive = interractive
        self.__args__ = (
            cmd,
            interractive,
        )

    def run(self, meta):
        if self.interractive:
            os.system(self.cmd)
        else:
            child = Process(self.cmd)
            events = multiprocessing.Queue()
            parent = multiprocessing.Process(name='task', target=child.run, args=(events, True))
            parent.start()
            exit = False
            returncode = None
            while not exit:
                try:
                    msg = events.get(timeout=0.1)
                except Empty:
                    if exit:
                        break
                else:
                    if msg.type == 'line':
                        print(term.lightblack('\u2502'), msg.data.decode('utf-8'), end='')
                    elif msg.type == 'start':
                        print('$ ' + term.lightwhite(self.cmd) + term.black('  # pid=%s' % msg.data['pid']))
                    elif msg.type == 'stop':
                        returncode = msg.data['returncode']
                        if returncode:
                            print(term.lightblack('\u2514' + term.red(' failed (rc={}). '.format(returncode))))
                        else:
                            print(term.lightblack('\u2514' + term.green(' success. ')))
                        exit = True
            if returncode:
                raise RuntimeError(
                    '"{command}" exited with status {returncode}.'.format(command=self.cmd, returncode=returncode)
                )
        self.set_complete()


class Make(System):
    def __init__(self, target):
        super().__init__('make ' + target)
        self.__args__ = (target, )


class Commit(Step):
    def __init__(self, message, *, tag=False):
        super().__init__()
        self.message = message
        self.tag = bool(tag)
        self.__args__ = (message, self.tag)

    def run(self, meta):
        branch = self.exec('git rev-parse --abbrev-ref HEAD')
        version = self.exec('python setup.py --version')
        assert version == meta['version']
        os.system('git commit -m ' + json.dumps(self.message.format(**meta)))
        if self.tag:
            os.system('git tag -am {version} {version}'.format(**meta))

        repo = Repo()
        for remote in repo.remotes:
            if str(remote) in ('origin', 'upstream'):
                self.logger.info('git push {} {}...'.format(remote, branch))
                os.system('git push {} {} --tags'.format(remote, branch))

        self.set_complete()



from medikit import steps


def setup_default_pipelines(config):
    # todo, move this in a configurable place
    with config.pipeline('release') as release:
        release.add(steps.Install())
        release.add(steps.BumpVersion())
        release.add(steps.Make('update-requirements'))
        release.add(steps.Make('clean install'))  # test docs
        release.add(steps.System('git add -p .', interractive=True))
        release.add(steps.Commit('Release: {version}', tag=True))



import sys
import warnings
from contextlib import contextmanager

from packaging import version

import medikit


@contextmanager
def deprecated_feature(deprecated_in, removed_in, feature, replacement):
    current_version = version.parse(medikit.__version__)
    deprecated_in = version.parse(deprecated_in)
    removed_in = version.parse(removed_in)
    print(current_version, deprecated_in, removed_in)
    print(sys.warnoptions)
    if current_version < deprecated_in:
        warnings.warn(
            'Using {} is pending deprecation starting at {} and will stop working at {}.\nPlease use {} instead.'.
            format(feature, deprecated_in, removed_in, replacement), PendingDeprecationWarning
        )
        yield
    elif current_version < removed_in:
        warnings.warn(
            'Using {} is deprecated since {} and will stop working at {}.\nPlease use {} instead.'.format(
                feature, deprecated_in, removed_in, replacement
            ), DeprecationWarning
        )
        yield
    else:
        raise RuntimeError(
            'Using {} is not supported anymore since {}. Use {} instead.'.format(feature, removed_in, replacement)
        )



"""
YAPF support, to automatically reformat all your (python) source code.

.. code-block:: shell-session

    $ make format

"""

import os

from medikit import settings
from medikit.events import subscribe
from medikit.feature import Feature, SUPPORT_PRIORITY, ABSOLUTE_PRIORITY


class YapfFeature(Feature):
    requires = {'python'}

    @subscribe('medikit.feature.python.on_generate')
    def on_python_generate(self, event):
        event.config['python'].add_requirements(dev=['yapf'])

    @subscribe('medikit.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile
        makefile['YAPF'] = '$(PYTHON) -m yapf'
        makefile['YAPF_OPTIONS'] = '-rip'
        makefile.add_target(
            'format',
            '''
            $(YAPF) $(YAPF_OPTIONS) .
            $(YAPF) $(YAPF_OPTIONS) Projectfile
        ''',
            deps=('install-dev', ),
            phony=True
        )

    @subscribe('medikit.on_start', priority=SUPPORT_PRIORITY)
    def on_start(self, event):
        self.render_file('.style.yapf', 'yapf/style.yapf.j2')

    @subscribe('medikit.on_start', priority=ABSOLUTE_PRIORITY - 1)
    def on_before_start(self, event):
        style_config = os.path.join(os.getcwd(), '.style.yapf')
        if os.path.exists(style_config):
            self.dispatcher.info('YAPF_STYLE_CONFIG = ' + style_config)
            settings.YAPF_STYLE_CONFIG = style_config


__feature__ = YapfFeature



"""
NodeJS / Yarn support.

This feature is experimental and as though it may work for you, that's not a guarantee. Please use with care.

"""

import json
import os
import runpy

from medikit.events import subscribe
from medikit.feature import Feature, LAST_PRIORITY
from medikit.feature.make import which
from medikit.steps.version import PythonVersion


class NodeJSConfig(Feature.Config):
    """ Configuration API for the «nodejs» feature. """

    def __init__(self):
        self._setup = {}
        self._dependencies = {
            None: {},
            'dev': {},
            'peer': {},
            'bundled': {},
            'optional': {},
        }
        self.base_dir = None

    def setup(self, *, base_dir=None):
        if base_dir:
            self.base_dir = base_dir
        return self

    def add_dependencies(self, deps=None, **kwargs):
        if deps:
            self.__add_dependencies(deps)
        for deptype, deps in kwargs.items():
            self.__add_dependencies(deps, deptype=deptype)
        return self

    def get_dependencies(self):
        return {
            deptype + 'Dependencies' if deptype else 'dependencies': deps
            for deptype, deps in self._dependencies.items() if len(deps)
        }

    def __add_dependencies(self, deps, deptype=None):
        if len(deps):
            if deptype not in self._dependencies:
                raise KeyError('Invalid dependency type ' + deptype)
        self._dependencies[deptype].update(deps)


class NodeJSFeature(Feature):
    requires = {'python', 'make'}

    Config = NodeJSConfig

    @subscribe('medikit.feature.make.on_generate')
    def on_make_generate(self, event):
        event.makefile['YARN'] = which('yarn')
        event.makefile['NODE'] = which('node')

        event.makefile.get_target('install').install += [
            '$(YARN) install --production',
        ]

        event.makefile.get_target('install-dev').install += [
            '$(YARN) install',
        ]

    @subscribe('medikit.on_start')
    def on_start(self, event):
        name = event.config['python'].get('name')

        current_version = PythonVersion.coerce(
            runpy.run_path(event.config['python'].version_file).get('__version__'), partial=True
        )
        current_version.partial = False

        package = {
            'name': name,
            'version': str(current_version),
            'description': event.config['python'].get('description'),
            'author': event.config['python'].get('author'),
            'license': event.config['python'].get('license'),
            **event.config['nodejs'].get_dependencies()
        }

        base_dir = event.config['nodejs'].base_dir or '.'

        self.render_file_inline(
            os.path.join(base_dir, 'package.json'),
            json.dumps(package, sort_keys=True, indent=4),
            override=True,
        )

    @subscribe('medikit.on_end', priority=LAST_PRIORITY)
    def on_end(self, event):
        base_dir = event.config['nodejs'].base_dir or '.'
        os.system('cd {base_dir}; yarn install'.format(base_dir=base_dir))
        os.system('cd {base_dir}; git add yarn.lock'.format(base_dir=base_dir))


__feature__ = NodeJSFeature



import contextlib
import os
import stat

from whistle import Event


class FileEvent(Event):
    def __init__(self, filename, executable, override):
        super(FileEvent, self).__init__()
        self.executable = executable
        self.filename = filename
        self.override = override


@contextlib.contextmanager
def File(dispatcher, name, *, executable=False, override=False):
    event = FileEvent(name, executable, override)

    if event.override or not os.path.exists(event.filename):
        with open(event.filename, 'w+') as f:
            event.file = f
            event = dispatcher.dispatch('medikit.on_file_opened', event)
            yield f
            event.file = None

        if event.executable and os.path.exists(event.filename):
            st = os.stat(event.filename)
            os.chmod(event.filename, st.st_mode | stat.S_IEXEC)

        dispatcher.dispatch('medikit.on_file_closed', event)
    else:
        with open('/dev/null', 'w') as f:
            yield f


@contextlib.contextmanager
def NullFile(dispatcher, name, *, executable=False, override=False):
    event = FileEvent(name, executable, override)

    if event.override or not os.path.exists(event.filename):
        with open('/dev/null', 'w') as f:
            event.file = f
            event = dispatcher.dispatch('medikit.on_file_opened', event)
            yield f
            event.file = None

        dispatcher.dispatch('medikit.on_file_closed', event)
    else:
        with open('/dev/null', 'w') as f:
            yield f



from stevedore import ExtensionManager

_all_features = {}


def load_feature_extensions():
    if not _all_features:

        def register_feature(ext, all_features=_all_features):
            all_features[ext.name] = ext.plugin

        mgr = ExtensionManager(namespace='medikit.feature')
        mgr.map(register_feature)

    return _all_features



import logging
import os
from contextlib import ContextDecorator

from jinja2 import Environment, PackageLoader, Template
from medikit import settings
from medikit.events import attach_subscriptions
from medikit.file import File
from medikit.settings import DEFAULT_FEATURES
from medikit.utils import is_identifier, format_file_content
from yapf import yapf_api

from mondrian import term

ABSOLUTE_PRIORITY = -100
HIGH_PRIORITY = -80
MEDIUM_PRIORITY = -60
LOW_PRIORITY = -60
SUPPORT_PRIORITY = -20
LAST_PRIORITY = 100


class Feature(object):
    _jinja_environment = None

    requires = set()
    conflicts = set()

    file_type = staticmethod(File)

    __usage__ = None

    class Config(ContextDecorator):
        def __enter__(self):
            return self

        def __exit__(self, *exc):
            return False

    def __init__(self, dispatcher):
        """

        :param LoggingDispatcher dispatcher:
        """
        self.dispatcher = dispatcher
        self.configure()
        attach_subscriptions(self, self.dispatcher)

    def configure(self):
        pass

    def file(self, *args, **kwargs):
        return self.file_type(self.dispatcher, *args, **kwargs)

    @property
    def __name__(self):
        return type(self).__name__

    @property
    def __shortname__(self):
        return self.__name__.replace('Feature', '').lower()

    @property
    def jinja(self):
        if type(self)._jinja_environment is None:
            type(self)._jinja_environment = Environment(loader=PackageLoader(__name__, 'template'))
        return type(self)._jinja_environment

    def _log_file(self, target, override, content=()):
        self.dispatcher.info(
            term.bold(term.red('W!') if override else term.green('W?')), target, '({} bytes)'.format(len(content))
        )

    def render(self, template, context=None):
        context = context or {}
        os.path.join(os.path.dirname(__file__), 'template')
        return self.jinja.get_template(template).render(**(context or {}))

    def render_file(self, target, template, context=None, *, executable=False, override=False, force_python=False):
        with self.file(target, executable=executable, override=override) as f:
            content = format_file_content(self.render(template, context))
            if force_python or target.endswith('.py'):
                content, modified = yapf_api.FormatCode(content, filename=target)
            f.write(content)
            self._log_file(target, override, content)

    def render_file_inline(self, target, template_string, context=None, override=False, force_python=False):
        with self.file(target, override=override) as f:
            content = format_file_content(Template(template_string).render(**(context or {})))
            if force_python or target.endswith('.py'):
                content, modified = yapf_api.FormatCode(
                    content, filename=target, style_config=settings.YAPF_STYLE_CONFIG
                )
            f.write(content)
            self._log_file(target, override, content)

    def render_empty_files(self, *targets, **kwargs):
        override = kwargs.pop('override', False)
        for target in targets:
            with self.file(target, override=override) as f:
                self._log_file(target, override)


class ProjectInitializer(Feature):
    def __init__(self, dispatcher, options):
        super().__init__(dispatcher)
        self.options = options

    def execute(self):
        context = {}

        if self.options.get('name'):
            if not is_identifier(self.options['name']):
                raise RuntimeError('Invalid package name {!r}.'.format(self.options['name']))
            context['name'] = self.options['name']
            logging.info('name = %s', context['name'])
        else:
            context['name'] = input('Name: ')
            while not is_identifier(context['name']):
                logging.error('Invalid name. Please only use valid python identifiers.')
                context['name'] = input('Name: ')

        if self.options.get('description'):
            context['description'] = self.options['description']
        else:
            context['description'] = input('Description: ')

        if self.options.get('license'):
            context['license'] = self.options['license']
        else:
            context['license'
                    ] = input('License [Apache License, Version 2.0]: ').strip() or 'Apache License, Version 2.0'

        context['url'] = ''
        context['download_url'] = ''
        context['author'] = ''
        context['author_email'] = ''

        context['features'] = DEFAULT_FEATURES
        if self.options.get('features'):
            context['features'] = context['features'].union(self.options['features'])

        context['requirements'] = self.options.get('requirements', [])

        self.render_file('Projectfile', 'Projectfile.j2', context, override=True, force_python=True)



"""
The «django» feature adds the django framework to your project.

"""

import os
import random

from medikit.events import subscribe

from . import Feature, SUPPORT_PRIORITY

random = random.SystemRandom()


def generate_secret_key():
    chars = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789!@#$%^&*(-_=+)'
    return ''.join(random.choice(chars) for i in range(64))


class DjangoConfig(Feature.Config):
    """ Configuration class for the “django” feature. """

    version = '~=2.0,<2.1'
    """Which django version requirement do you want?"""

    __usage__ = """
    
    This will add a few items to your Makefile, and ensure a minimalistic django project structure is available.
    
    By default, it will use Django {version}, but you can tune that:
    
    .. code-block:: python
    
        django.version = '~=2.0.3'
        
    This feature will also add or extend a "prod" python extra, that will install gunicorn.
    
    """.format(version=version)

    def __init__(self):
        self._static_dir = None
        self.version = self.version

    @property
    def static_dir(self):
        """
        The django global static directory (property).
        """
        return self._static_dir

    @static_dir.setter
    def static_dir(self, value):
        self._static_dir = value


class DjangoFeature(Feature):
    requires = {'python'}

    Config = DjangoConfig

    @subscribe('medikit.feature.python.on_generate')
    def on_python_generate(self, event):
        event.config['python'].add_requirements(
            'django ' + event.config['django'].version,
            prod=[
                'gunicorn ==19.7.1',
            ],
        )

    @subscribe('medikit.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile
        makefile['DJANGO'] = '$(PYTHON) manage.py'
        makefile.add_target('runserver', '''$(DJANGO) runserver''', deps=('install-dev', ), phony=True)

    @subscribe('medikit.on_start')
    def on_start(self, event):
        django_config = event.config['django']
        python_config = event.config['python']

        context = {
            **python_config.get_setup(),
            'config_package': 'config',
            'secret_key': generate_secret_key(),
        }

        # Create configuration
        config_path = 'config'
        if not os.path.exists(config_path):
            os.makedirs(config_path)

        self.render_file('manage.py', 'django/manage.py.j2', context, executable=True, force_python=True)
        self.render_file(os.path.join(config_path, 'settings.py'), 'django/settings.py.j2', context, force_python=True)
        self.render_file(os.path.join(config_path, 'urls.py'), 'django/urls.py.j2', context, force_python=True)
        self.render_file(os.path.join(config_path, 'wsgi.py'), 'django/wsgi.py.j2', context, force_python=True)

        self.dispatcher.dispatch('medikit.feature.django.on_configure')

        static_dir = django_config.static_dir or os.path.join(python_config.package_dir, 'static')
        if not os.path.exists(static_dir):
            os.makedirs(static_dir)
        self.render_empty_files(os.path.join(static_dir, 'favicon.ico'))


__feature__ = DjangoFeature



"""
Adds docker capabilities to your package, using either "docker build" or "rocker build" to create an image containing
your code, in a fully functionnal python virtualenv.

"""
from argparse import Namespace

from medikit.events import subscribe
from medikit.feature import Feature
from medikit.feature.make import which
from medikit.structs import Script

DEFAULT_NAME = '$(shell echo $(PACKAGE) | tr A-Z a-z)'

DOCKER = 'docker'
ROCKER = 'rocker'


class DockerConfig(Feature.Config):
    __usage__ = """
    
    *This feature is brand new and should be used with care.*
    
    You'll get a few make targets out of this, and a Dockerfile (or Rockerfile).
    
    Build
    -----
    
    Building an image is as simple as running:
    
    .. code-block:: shell-session
    
        $ make docker-build
        
    This will use the root Dockerfile (or Rockerfile, see builders below) and build an image named after your package.
    
    Push
    ----
    
    You can push the built image to a docker registry, using:
    
    .. code-block:: shell-session
    
        $ make docker-push
        
    Run
    ---
    
    You can run the default entrypoint / command in a new container using the built image:
    
    .. code-block:: shell-session
    
        $ make docker-run
        
    Shell
    -----
    
    You can run a bash shell in a new container using the built image:
    
    .. code-block:: shell-session
    
        $ make docker-shell
    
    Custom builder
    --------------
    
    You can change the default docker builder (a.k.a "docker build") to use rocker
    (see https://github.com/grammarly/rocker).
    
    .. code-block:: python
    
        docker.use_rocker_builder()
        
    Custom image name or registry
    -----------------------------
    
    If you want to customize the image name, or the target registry (for example if you want to use Amazon Elastic 
    Container Registry, Google Container Registry, Quay, or even a private registry you're crazy enough to host
    yourself):
    
    .. code-block:: python
    
        # only override the registry
        docker.set_remote(registry='eu.gcr.io')
        
        # override the registry and username, but keep the default image name
        docker.set_remote(registry='eu.gcr.io', user='sergey')
        
        # override the image name only
        docker.set_remote(name='acme')
        
    Docker Compose
    --------------
    
    The feature will also create an example docker-compose.yml file.
    
    If you don't want this, you can:
    
    .. code-block:: python
    
        docker.compose_file = None
        
    Or if you want to override its name:
    
    .. code-block:: python
    
        docker.compose_file = 'config/docker/compose.yml'
        
    Please note that this file will only contain a structure skeleton, with no service defined. This is up to you to
    fill, although we may work on this in the future as an opt-in managed file.
    
    """

    def __init__(self):
        self._registry = None
        self._user = None
        self._name = DEFAULT_NAME

        self._build_file = None
        self._compose_file = None

        self.use_default_builder()

    def set_remote(self, registry=None, user=None, name=DEFAULT_NAME):
        self._registry = registry
        self._user = user
        self._name = name

    @property
    def compose_file(self):
        if self._compose_file is None:
            return 'docker-compose.yml'
        if self._compose_file is False:
            return None
        return self._compose_file

    @compose_file.setter
    def compose_file(self, value):
        self._compose_file = value

    @property
    def build_file(self):
        if self._build_file is None:
            return self.builder.title() + 'file'
        return self._build_file

    @build_file.setter
    def build_file(self, value):
        self._build_file = value

    def _get_default_variables(self):
        return dict(
            DOCKER=which('docker'),
            DOCKER_BUILD='$(DOCKER) image build',
            DOCKER_BUILD_OPTIONS='--build-arg IMAGE=$(DOCKER_IMAGE) --build-arg TAG=$(DOCKER_TAG)',
            DOCKER_PUSH='$(DOCKER) image push',
            DOCKER_PUSH_OPTIONS='',
            DOCKER_RUN='$(DOCKER) run',
            DOCKER_RUN_COMMAND='',
            DOCKER_RUN_OPTIONS='',
        )

    def _get_default_image_variables(self):
        return dict(
            DOCKER_BUILD_FILE='',  # will be set at runtime.
            DOCKER_IMAGE='',  # will be set at runtime, see #71.
            DOCKER_TAG='$(VERSION)',
        )

    def use_default_builder(self):
        self.builder = DOCKER

        self._variables = [
            self._get_default_variables(),
            self._get_default_image_variables(),
        ]

        self.scripts = Namespace(
            build=Script(
                '$(DOCKER_BUILD) -f $(DOCKER_BUILD_FILE) $(DOCKER_BUILD_OPTIONS) -t $(DOCKER_IMAGE):$(DOCKER_TAG) .'),
            push=Script('$(DOCKER_PUSH) $(DOCKER_PUSH_OPTIONS) $(DOCKER_IMAGE):$(DOCKER_TAG)'),
            run=Script(
                '$(DOCKER_RUN) $(DOCKER_RUN_OPTIONS) --interactive --tty --rm --name=$(PACKAGE)_run -p 8000:8000 $(DOCKER_IMAGE):$(DOCKER_TAG) $(DOCKER_RUN_COMMAND)'
            ),
            shell=Script('DOCKER_RUN_COMMAND="/bin/bash" $(MAKE) docker-run'),
        )

    def use_rocker_builder(self):
        self.builder = ROCKER

        self._variables = [
            self._get_default_variables(),
            self._get_default_image_variables(),
            dict(
                ROCKER=which('rocker'),
                ROCKER_BUILD='$(ROCKER) build',
                ROCKER_BUILD_OPTIONS='',
                ROCKER_BUILD_VARIABLES=
                '--var DOCKER_IMAGE=$(DOCKER_IMAGE) --var DOCKER_TAG=$(DOCKER_TAG) --var PYTHON_REQUIREMENTS_FILE=requirements-prod.txt',
            ),
        ]

        self.scripts.build.set('$(ROCKER_BUILD) $(ROCKER_BUILD_OPTIONS) $(ROCKER_BUILD_VARIABLES) .')
        self.scripts.push.set('ROCKER_BUILD_OPTIONS="$(ROCKER_BUILD_OPTIONS) --push" $(MAKE) docker-build')

    @property
    def variables(self):
        for variables in self._variables:
            yield from sorted(variables.items())

    @property
    def image(self):
        return '/'.join(filter(None, (self._registry, self._user, self._name)))


class DockerFeature(Feature):
    Config = DockerConfig

    @subscribe('medikit.feature.make.on_generate', priority=-1)
    def on_make_generate(self, event):
        docker_config = event.config['docker']

        for var, val in docker_config.variables:
            event.makefile[var] = val

        # Set DOCKER_IMAGE at runtime, see #71.
        if not event.makefile['DOCKER_BUILD_FILE']:
            event.makefile['DOCKER_BUILD_FILE'] = docker_config.build_file
        if not event.makefile['DOCKER_IMAGE']:
            event.makefile['DOCKER_IMAGE'] = docker_config.image

        # Targets
        event.makefile.add_target('docker-build', docker_config.scripts.build, phony=True, doc='Build a docker image.')
        event.makefile.add_target(
            'docker-push', docker_config.scripts.push, phony=True, doc='Push docker image to remote registry.'
        )
        event.makefile.add_target(
            'docker-run',
            docker_config.scripts.run,
            phony=True,
            doc='Run the default entry point in a container based on our docker image.'
        )
        event.makefile.add_target(
            'docker-shell',
            docker_config.scripts.shell,
            phony=True,
            doc='Run bash in a container based on our docker image.'
        )

    @subscribe('medikit.on_end')
    def on_end(self, event):
        docker_config = event.config['docker']

        self.render_file_inline(
            '.dockerignore', '''
            **/__pycache__
            *.egg-info
            .cache
            .git
            .idea
            /Dockerfile
            /Projectfile
            /Rockerfile
            node_modules
            static
        ''', event.variables
        )

        if docker_config.compose_file:
            self.render_file_inline(
                docker_config.compose_file, '''
                version: '3'

                volumes:
                
                #   postgres_data: {}

                services:
                
                #   postgres:
                #     image: postgres:10
                #     ports:
                #       - 5432:5432
                #     volumes:
                #       - postgres_data:/var/lib/postgresql/data
            '''
            )

        if docker_config.builder == DOCKER:
            self.render_file_inline(docker_config.build_file, '''
                FROM python:3
            ''')
        elif docker_config.builder == ROCKER:
            self.render_file_inline(
                docker_config.build_file, '''
                FROM python:3
                 
                # Mount cache volume to keep cache persistent from one build to another
                MOUNT /app/.cache
                WORKDIR /app
                
                # Create application user
                RUN useradd --home-dir /app --group www-data app \
                 && pip install -U pip wheel virtualenv \
                 && mkdir /env \
                 && chown app:www-data -R /app /env 
                 
                # Add and install python requirements in a virtualenv
                USER app
                RUN virtualenv -p python3 /env/
                ADD setup.py *.txt /app/
                RUN /env/bin/pip install -r {{ '{{ .PYTHON_REQUIREMENTS_FILE }}' }}

                # Add everything else
                USER root
                ADD . /app
                # IMPORT /static /app
                # IMPORT /assets.json /app
                RUN chown app:www-data -R /app

                # Entrypoint
                USER app
                CMD /env/bin/gunicorn config.wsgi --bind 0.0.0.0:8000 --workers 4
                
                PUSH {{ '{{ .DOCKER_IMAGE }}:{{ .DOCKER_TAG }}' }}
            '''
            )
        else:
            raise NotImplementedError('Unknown builder {}'.format(docker_config.builder))


__feature__ = DockerFeature



import datetime
import os

from mondrian import term

from medikit.commands.base import Command
from medikit.commands.utils import _read_configuration
from medikit.events import LoggingDispatcher
from medikit.pipeline import ConfiguredPipeline, logger

START = 'start'
CONTINUE = 'continue'
ABORT = 'abort'
COMPLETE = 'complete'


class PipelineCommand(Command):
    def add_arguments(self, parser):
        parser.add_argument('pipeline', default=None, nargs='?')
        parser.add_argument(
            'action', choices=(
                START,
                CONTINUE,
                ABORT,
            ), nargs='?'
        )
        parser.add_argument('--force', '-f', action='store_true')

    @classmethod
    def handle(cls, config_filename, *, pipeline, action, force=False, verbose=False):
        dispatcher = LoggingDispatcher()
        variables, features, files, config = _read_configuration(dispatcher, config_filename)

        if not pipeline:
            raise ValueError(
                'You must choose a pipeline to run. Available choices: {}.'.format(
                    ', '.join(sorted(config.pipelines.keys()))
                )
            )

        if not pipeline in config.pipelines:
            raise ValueError(
                'Undefined pipeline {!r}. Valid choices are: {}.'.format(
                    pipeline, ', '.join(sorted(config.pipelines.keys()))
                )
            )

        pipeline = ConfiguredPipeline(pipeline, config.pipelines[pipeline])
        path = os.path.dirname(config_filename)
        pipeline_file = os.path.join(path, '.medikit/pipelines', pipeline.name + '.json')
        pipeline_dirname = os.path.dirname(pipeline_file)
        if not os.path.exists(pipeline_dirname):
            os.makedirs(pipeline_dirname)
        elif not os.path.isdir(pipeline_dirname):
            raise NotADirectoryError(
                'The pipeline state path "{}" was found but is not a directory...'.format(pipeline_dirname)
            )

        if not action:
            raise RuntimeError('Choose a pipeline action: start, continue, abort.')

        while action:
            if action == START:
                action = cls._handle_start(pipeline, filename=pipeline_file, force=force)
            elif action == CONTINUE:
                action = cls._handle_continue(pipeline, filename=pipeline_file)
            elif action == ABORT:
                action = cls._handle_abort(pipeline, filename=pipeline_file)
            elif action == COMPLETE:
                target = os.path.join(
                    '.medikit/pipelines',
                    pipeline.name + '.' + str(datetime.datetime.now()).replace(':', '.').replace(' ', '.') + '.json'
                )
                os.rename(pipeline_file, os.path.join(path, target))
                logger.info('Pipeline complete. State saved as “{}”.'.format(target))
                break
            else:
                raise ValueError('Invalid action “{}”.'.format(action))
            force = False

    @classmethod
    def _handle_start(cls, pipeline, *, filename, force=False):
        if os.path.exists(filename):
            if not force:
                raise FileExistsError(
                    'Already started, use `medikit pipeline {name} start --force` to force a restart, or use `medikit pipeline {name} continue`.'.
                    format(name=pipeline.name)
                )
            os.unlink(filename)

        # Initialize pipeline state to "just started".
        pipeline.init()

        # Write the new, empty state file
        with open(filename, 'w+') as f:
            f.write(pipeline.serialize())

        # "Continue", until step 1.
        return CONTINUE

    @classmethod
    def _handle_continue(cls, pipeline, *, filename):
        if not os.path.exists(filename):
            raise FileNotFoundError(
                'Pipeline “{}” not started, hence you cannot “continue” it. Are you looking for `medikit pipeline {name} start`?'.
                format(name=pipeline.name)
            )

        # XXX TODO add a lock file during the step and unlock at the end.
        with open(filename) as f:
            pipeline.unserialize(f.read())

        try:
            step = pipeline.next()
            name, current, size, descr = pipeline.name, pipeline.current, len(pipeline), str(step)
            logger.info(
                term.black(' » ').join(
                    (
                        term.lightblue('{} ({}/{})'.format(name.upper(), current, size)),
                        term.yellow('⇩ BEGIN'),
                        term.lightblack(descr),
                    )
                )
            )
            step.run(pipeline.meta)
            if step.complete:
                logger.info(
                    term.black(' » ')
                    .join((
                        term.lightblue('{} ({}/{})'.format(name.upper(), current, size)),
                        term.green('SUCCESS'),
                    )) + '\n'
                )
            else:
                logger.info(
                    term.black(' » ')
                    .join((
                        term.lightblue('{} ({}/{})'.format(name.upper(), current, size)),
                        term.red('FAILED'),
                    )) + '\n'
                )
                return

        except StopIteration:
            return COMPLETE

        with open(filename, 'w+') as f:
            f.write(pipeline.serialize())

        return CONTINUE

    @classmethod
    def _handle_abort(cls, pipeline, *, filename):
        assert os.path.exists(filename)
        try:
            with open(filename) as f:
                pipeline.unserialize(f.read())
            pipeline.abort()
        finally:
            os.unlink(filename)



import sys

from medikit.steps.exec import System


class Install(System):
    def __init__(self, *packages):
        if not len(packages):
            packages = ('pip', 'wheel', 'twine')
        packages = list(sorted(packages))
        super().__init__(sys.executable + ' -m pip install --upgrade ' + ' '.join(packages))
        self.__args__ = (*packages, )



from medikit.steps.base import Step
from medikit.steps.install import Install
from medikit.steps.version import BumpVersion
from medikit.steps.exec import Make, System, Commit

__all__ = [
    'BumpVersion',
    'Install',
    'Make',
    'Commit',
    'System',
    'Step',
]



import runpy

import medikit
from medikit.compat import deprecated_feature
from medikit.config.defaults import setup_default_pipelines
from medikit.config.registry import ConfigurationRegistry
from medikit.settings import DEFAULT_FEATURES
from medikit.utils import format_file_content


def read_configuration(dispatcher, filename, variables, features, files):
    config = ConfigurationRegistry()
    setup_default_pipelines(config)
    default_context = {'listen': dispatcher.listen}

    # monkey patch placeholders
    _listen, _pipeline, _require = medikit.listen, medikit.pipeline, medikit.require
    try:
        medikit.listen, medikit.pipeline, medikit.require = dispatcher.listen, config.pipeline, config.require
        context = runpy.run_path(filename, init_globals=default_context)
    finally:
        # restore old values
        medikit.listen, medikit.pipeline, medikit.require = _listen, _pipeline, _require

    for k in variables.keys():
        if k in context:
            variables[k] = context[k]

    for feature in DEFAULT_FEATURES:
        config.require(feature)

    # Deprecated: enabled and disabled features.
    enable_features, disable_features = context.pop('enable_features', ()), context.pop('disable_features', ())
    if len(enable_features) or len(disable_features):
        with deprecated_feature('0.5.0', '0.6.0', 'Using "enable_features" and "disable_features"', 'require()'):
            for feature in set(enable_features) - set(disable_features):
                config.require(feature)

    # Better: required features.
    features = features | set(config.keys())

    for k in files.keys():
        if k in context:
            files[k] = format_file_content(context[k])

    return variables, features, files, config



from pip._vendor.distlib.util import parse_requirement

from medikit.events import subscribe
from . import Feature, SUPPORT_PRIORITY


class SphinxConfig(Feature.Config):
    __usage__ = """
    
        You can customize the theme:
        
        .. code-block:: python
        
            sphinx.theme = 'sphinx_rtd_theme'
            
        Note that this should be a parsable requirement, and it won't be added to your docs/conf.py automatically.
        
        This feature will add the necessary requirements to the python feature (sphinx mostly, and eventually your theme requirement) and setup the correct makefile task to build the html docs. Note that it won't bootstrap your sphinx config file, and you still need to run the following if your documentation does not exist yet:
        
        .. code-block:: shell-session
        
            $ make update-requirements
            $ make install-dev
            $ mkdir docs
            $ cd docs
            $ sphinx-quickstart .
            
        Then, eventually tune the configuration.
        
        .. note::
        
            In the future, we may consider generating it for you if it does not exist, but it's not a priority.
    
    """

    def __init__(self):
        self._theme = None

    def set_theme(self, theme):
        """
        Sets the theme. Prefer using the .theme property.

        :param theme: Requirement for theme
        """
        self._theme = parse_requirement(theme)

    def get_theme(self):
        """
        Gets the theme. Prefer using the .theme property.

        :return: parsed requirement
        """
        return self._theme

    theme = property(
        fget=get_theme, fset=set_theme, doc="Sphinx theme to use, that should be parsable as a requirement."
    )


class SphinxFeature(Feature):
    Config = SphinxConfig

    @subscribe('medikit.feature.python.on_generate')
    def on_python_generate(self, event):
        event.config['python'].add_requirements(dev=[
            'sphinx ~=1.7',
        ])

        theme = event.config['sphinx'].get_theme()
        if theme:
            event.config['python'].add_requirements(dev=[theme.requirement])

    @subscribe('medikit.feature.make.on_generate', priority=SUPPORT_PRIORITY)
    def on_make_generate(self, event):
        makefile = event.makefile

        makefile['SPHINX_BUILD'] = '$(PYTHON_DIRNAME)/sphinx-build'
        makefile['SPHINX_OPTIONS'] = ''
        makefile['SPHINX_SOURCEDIR'] = 'docs'
        makefile['SPHINX_BUILDDIR'] = '$(SPHINX_SOURCEDIR)/_build'

        makefile.add_target(
            '$(SPHINX_SOURCEDIR)',
            '''
            $(SPHINX_BUILD) -b html -D latex_paper_size=a4 $(SPHINX_OPTIONS) $(SPHINX_SOURCEDIR) $(SPHINX_BUILDDIR)/html
        ''',
            deps=('install-dev', ),
            phony=True
        )


__feature__ = SphinxFeature



# Generated by Medikit 0.5.11 on 2018-03-16.
# All changes will be overriden.
# Edit Projectfile and run “make update” (or “medikit update”) to regenerate.

from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Py3 compatibility hacks, borrowed from IPython.
try:
    execfile
except NameError:

    def execfile(fname, globs, locs=None):
        locs = locs or globs
        exec(compile(open(fname).read(), fname, "exec"), globs, locs)


# Get the long description from the README file
try:
    with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = f.read()
except:
    long_description = ''

# Get the classifiers from the classifiers file
tolines = lambda c: list(filter(None, map(lambda s: s.strip(), c.split('\n'))))
try:
    with open(path.join(here, 'classifiers.txt'), encoding='utf-8') as f:
        classifiers = tolines(f.read())
except:
    classifiers = []

version_ns = {}
try:
    execfile(path.join(here, 'medikit/_version.py'), version_ns)
except EnvironmentError:
    version = 'dev'
else:
    version = version_ns.get('__version__', 'dev')

setup(
    author='Romain Dorgueil',
    author_email='romain@dorgueil.net',
    description='Opinionated python 3.5+ project management.',
    license='Apache License, Version 2.0',
    name='medikit',
    python_requires='>=3.5',
    version=version,
    long_description=long_description,
    classifiers=classifiers,
    packages=find_packages(exclude=['ez_setup', 'example', 'test']),
    include_package_data=True,
    install_requires=[
        'git-semver (~= 0.2.1)', 'jinja2 (~= 2.9)', 'mondrian (~= 0.6)', 'packaging (~= 17.0)', 'pip-tools (~= 1.11)',
        'stevedore (~= 1.28)', 'whistle (~= 1.0)', 'yapf (~= 0.20)'
    ],
    extras_require={
        'dev': [
            'coverage (~= 4.4)', 'pytest (~= 3.4)', 'pytest-cov (~= 2.5)', 'pytest-sugar (~= 0.9.1)', 'sphinx (~= 1.7)',
            'sphinx-sitemap (>= 0.2, < 0.3)', 'yapf'
        ]
    },
    entry_points={
        'console_scripts': ['medikit=medikit.__main__:main'],
        'medikit.feature': [
            'django = medikit.feature.django:DjangoFeature', 'docker = medikit.feature.docker:DockerFeature',
            'git = medikit.feature.git:GitFeature', 'make = medikit.feature.make:MakeFeature',
            'nodejs = medikit.feature.nodejs:NodeJSFeature', 'pylint = medikit.feature.pylint:PylintFeature',
            'pytest = medikit.feature.pytest:PytestFeature', 'python = medikit.feature.python:PythonFeature',
            'sphinx = medikit.feature.sphinx:SphinxFeature', 'webpack = medikit.feature.webpack:WebpackFeature',
            'yapf = medikit.feature.yapf:YapfFeature'
        ]
    },
    url='https://python-medikit.github.io/',
    download_url='https://github.com/python-medikit/medikit/archive/{version}.tar.gz'.format(version=version),
)



#!/usr/bin/env python

import logging
import os
import sys
import warnings
from subprocess import check_output

import mondrian

import medikit
from medikit.commands.main import MedikitCommand


def main(args=None):
    if not sys.warnoptions:
        logging.captureWarnings(True)
    warnings.simplefilter("default", DeprecationWarning)
    mondrian.setup(excepthook=True)
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    cli = MedikitCommand()

    options, more_args = cli.parser.parse_known_args(args if args is not None else sys.argv[1:])
    if options.verbose:
        logger.setLevel(logging.DEBUG)

    options = vars(options)
    command, handler = options.pop('command'), options.pop('_handler')

    config_filename = os.path.join(os.getcwd(), options.pop('target', '.'), options.pop('config'))

    version = medikit.__version__
    try:
        if os.path.exists(os.path.join(os.path.dirname(os.path.dirname(medikit.__file__)), '.git')):
            try:
                version = check_output(
                    ['git', 'describe'], cwd=os.path.dirname(os.path.dirname(medikit.__file__))
                ).decode('utf-8').strip() + ' (git)'
            except:
                version = check_output(
                    ['git', 'rev-parse', 'HEAD'], cwd=os.path.dirname(os.path.dirname(medikit.__file__))
                ).decode('utf-8').strip()[0:7] + ' (git)'
    except:
        warnings.warn('Git repository found, but could not find version number from the repository.')

    print(mondrian.term.lightwhite_bg(mondrian.term.red('  ✚  Medikit v.' + version + '  ✚  ')))

    if len(more_args):
        return handler(config_filename, more=more_args, **options)
    else:
        return handler(config_filename, **options)


if __name__ == '__main__':
    main()



from unittest import TestCase

from medikit.config import ConfigurationRegistry
from medikit.events import LoggingDispatcher
from medikit.feature import Feature
from medikit.file import NullFile


class FeatureTestCase(TestCase):
    feature_type = Feature
    required_features = set()

    def create_dispatcher(self):
        return LoggingDispatcher()

    def create_feature(self, feature_type=None, dispatcher=None):
        dispatcher = dispatcher or self.create_dispatcher()
        feature = (feature_type or self.feature_type)(dispatcher)
        feature.file_type = NullFile
        return feature, dispatcher

    def create_config(self):
        config = ConfigurationRegistry()
        if len(self.required_features):
            config.require(*self.required_features)
        return config



from medikit.commands.init import InitCommand
from medikit.commands.pipeline import PipelineCommand
from medikit.commands.update import UpdateCommand

# XXX deprecated aliases
handle_init = InitCommand.handle
handle_pipeline = PipelineCommand.handle
handle_update = UpdateCommand.handle



from medikit._version import __version__


def listen(*args, **kwargs):
    raise NotImplementedError('This is a stub.')


def require(*args):
    raise NotImplementedError('This is a stub.')


def pipeline(*args):
    raise NotImplementedError('This is a stub.')


__all__ = [
    '__version__',
]



import datetime
import keyword
import textwrap
import tokenize

import medikit


def is_identifier(ident):
    """Determines, if string is valid Python identifier."""

    # Smoke test — if it's not string, then it's not identifier, but we don't
    # want to just silence exception. It's better to fail fast.
    if not isinstance(ident, str):
        raise TypeError('expected str, but got {!r}'.format(type(ident)))

    # Quick test — if string is in keyword list, it's definitely not an ident.
    if keyword.iskeyword(ident):
        return False

    readline = (lambda: (yield ident.encode('utf-8-sig')))().__next__
    tokens = list(tokenize.tokenize(readline))

    # You should get exactly 3 tokens
    if len(tokens) != 3:
        return False

    # First one is ENCODING, it's always utf-8 because we explicitly passed in
    # UTF-8 BOM with ident.
    if tokens[0].type != tokenize.ENCODING:
        return False

    # Second is NAME, identifier.
    if tokens[1].type != tokenize.NAME:
        return False

    # Name should span all the string, so there would be no whitespace.
    if ident != tokens[1].string:
        return False

    # Third is ENDMARKER, ending stream
    if tokens[2].type != tokenize.ENDMARKER:
        return False

    return True


def format_file_content(s):
    return textwrap.dedent(s).strip() + '\n'


def get_override_warning_banner(*, prefix='# ', above=None, bellow=None):
    return '\n'.join(
        filter(
            None, (
                above,
                textwrap.indent(
                    '\n'.join(
                        (
                            'Generated by Medikit ' + medikit.__version__ + ' on ' +
                            str(datetime.datetime.now().date()) + '.',
                            'All changes will be overriden.',
                            'Edit Projectfile and run “make update” (or “medikit update”) to regenerate.',
                        )
                    ),
                    prefix=prefix
                ), bellow
            )
        )
    )



from medikit.commands import InitCommand, UpdateCommand, PipelineCommand
from medikit.commands.base import Command


class MedikitCommand(Command):
    def add_arguments(self, parser):
        parser.add_argument('--config', '-c', default='Projectfile')
        parser.add_argument('--verbose', '-v', action='store_true', default=False)

        with self.create_child('command', required=True) as actions:
            # todo aliases for update/init
            # warning: http://bugs.python.org/issue9234
            actions.add('init', InitCommand, help='Create an empty project.')
            actions.add('update', UpdateCommand, help='Update current project.')
            actions.add('pipeline', PipelineCommand, help='Execute multi-steps pipelines (release, etc.).')